@@ -1,6 +1,8 @@
 -- DAV.hs: WebDAV client library
 -- Copyright © 2012  Clint Adams
 --
+-- vim: softtabstop=4:shiftwidth=4:expandtab
+--
 -- This program is free software: you can redistribute it and/or modify
 -- it under the terms of the GNU General Public License as published by
 -- the Free Software Foundation, either version 3 of the License, or
@@ -19,15 +21,20 @@
 module Network.Protocol.HTTP.DAV (
     DAVState
   , DAVContext(..)
+  , getProps
   , getPropsAndContent
   , putContentAndProps
+  , putContent
+  , deleteContent
+  , moveContent
+  , makeCollection
   , module Network.Protocol.HTTP.DAV.TH
 ) where
 
 import Network.Protocol.HTTP.DAV.TH
 
 import Control.Applicative (liftA2)
-import Control.Exception.Lifted (catchJust, finally)
+import Control.Exception.Lifted (catchJust, finally, bracketOnError)
 import Control.Lens ((.~), (^.))
 import Control.Monad (when)
 import Control.Monad.Trans (lift)
@@ -36,11 +43,12 @@ import Control.Monad.Trans.State.Lazy (evalStateT, StateT, get, modify)
 
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Lazy as BL
+import qualified Data.Map as Map
 
 import Data.Maybe (fromMaybe)
 
 import Network.HTTP.Conduit (httpLbs, parseUrl, applyBasicAuth, Request(..), RequestBody(..), Response(..), newManager, closeManager, ManagerSettings(..), def, HttpException(..))
-import Network.HTTP.Types (headerContentType, Method, RequestHeaders, unauthorized401)
+import Network.HTTP.Types (hContentType, Method, Status, RequestHeaders, unauthorized401, conflict409)
 
 import qualified Text.XML as XML
 import Text.XML.Cursor (($/), (&/), element, node, fromDocument, checkName)
@@ -69,14 +77,16 @@ davRequest meth addlhdrs rbody = do
     ctx <- get
     let req = (ctx ^. baseRequest) { method = meth, requestHeaders = (mk "User-Agent", "hDav 9.0"):addlhdrs, requestBody = rbody }
     let authreq = applyBasicAuth (ctx ^. basicusername) (ctx ^. basicpassword) req
-    resp <- lift (catchJust (is401Exception)
+    resp <- lift (catchJust (matchStatusCodeException unauthorized401)
                             (httpLbs req (ctx ^. httpManager))
                             (\_ -> httpLbs authreq (ctx ^. httpManager)))
     return resp
 
-is401Exception :: HttpException -> Maybe ()
-is401Exception (StatusCodeException s _) = if s == unauthorized401 then Just () else Nothing
-is401Exception _ = Nothing
+matchStatusCodeException :: Status -> HttpException -> Maybe ()
+matchStatusCodeException want (StatusCodeException s _)
+    | s == want = Just ()
+    | otherwise = Nothing
+matchStatusCodeException _ _ = Nothing
 
 emptyBody :: RequestBody m
 emptyBody = RequestBodyLBS BL.empty
@@ -94,7 +104,7 @@ getOptions = do
 
 lockResource :: MonadResourceBase m => Bool -> DAVState m ()
 lockResource nocreate = do
-    let ahs' = [headerContentType "application/xml; charset=\"utf-8\"", (mk "Depth", "0"), (mk "Timeout", "Second-300")]
+    let ahs' = [(hContentType, "application/xml; charset=\"utf-8\""), (mk "Depth", "0"), (mk "Timeout", "Second-300")]
     let ahs = if nocreate then (mk "If-Match", "*"):ahs' else ahs'
     lockresp <- davRequest "LOCK" ahs (xmlBody locky)
     let hdrtoken = (lookup "Lock-Token" . responseHeaders) lockresp
@@ -112,33 +122,49 @@ unlockResource = do
 supportsLocking :: DAVContext a -> Bool
 supportsLocking = liftA2 (&&) ("LOCK" `elem`) ("UNLOCK" `elem`) . _allowedMethods
 
-getAllProps :: MonadResourceBase m => DAVState m XML.Document
-getAllProps = do
-    let ahs = [headerContentType "application/xml; charset=\"utf-8\""]
+getPropsM :: MonadResourceBase m => DAVState m XML.Document
+getPropsM = do
+    let ahs = [(hContentType, "application/xml; charset=\"utf-8\"")]
     propresp <- davRequest "PROPFIND" ahs (xmlBody propname)
     return $ (XML.parseLBS_ def . responseBody) propresp
 
-getContent :: MonadResourceBase m => DAVState m (Maybe B.ByteString, BL.ByteString)
-getContent = do
+getContentM :: MonadResourceBase m => DAVState m (Maybe B.ByteString, BL.ByteString)
+getContentM = do
     resp <- davRequest "GET" [] emptyBody
-    let ct = lookup (mk "Content-Type") (responseHeaders resp)
+    let ct = lookup (hContentType) (responseHeaders resp)
     return $ (ct, responseBody resp)
 
-putContent :: MonadResourceBase m => (Maybe B.ByteString, BL.ByteString) -> DAVState m ()
-putContent (ct, body) = do
+putContentM :: MonadResourceBase m => (Maybe B.ByteString, BL.ByteString) -> DAVState m ()
+putContentM (ct, body) = do
     d <- get
     let ahs' = fromMaybe [] (fmap (return . (,) (mk "If") . parenthesize) (d ^. lockToken))
-    let ahs = ahs' ++ fromMaybe [] (fmap (return . (,) (mk "Content-Type")) ct)
+    let ahs = ahs' ++ fromMaybe [] (fmap (return . (,) (hContentType)) ct)
     _ <- davRequest "PUT" ahs (RequestBodyLBS body)
     return ()
 
+delContentM :: MonadResourceBase m => DAVState m ()
+delContentM = do
+    _ <- davRequest "DELETE" [] emptyBody
+    return ()
+
+moveContentM :: MonadResourceBase m => B.ByteString -> DAVState m ()
+moveContentM newurl = do
+    let ahs = [ (mk "Destination", newurl) ]
+    _ <- davRequest "MOVE" ahs emptyBody
+    return ()
+
+mkCol :: MonadResourceBase m => DAVState m ()
+mkCol = do
+    _ <- davRequest "MKCOL" [] emptyBody
+    return ()
+
 parenthesize :: B.ByteString -> B.ByteString
 parenthesize x = B.concat ["(", x, ")"]
 
-putProps :: MonadResourceBase m => XML.Document -> DAVState m ()
-putProps props = do
+putPropsM :: MonadResourceBase m => XML.Document -> DAVState m ()
+putPropsM props = do
     d <- get
-    let ah' = headerContentType "application/xml; charset=\"utf-8\""
+    let ah' = (hContentType, "application/xml; charset=\"utf-8\"")
     let ahs = ah':fromMaybe [] (fmap (return . (,) (mk "If") . parenthesize) (_lockToken d))
     _ <- davRequest "PROPPATCH" ahs ((RequestBodyLBS . props2patch) props) -- FIXME: should diff and remove props from target
     return ()
@@ -146,41 +172,88 @@ putProps props = do
 props2patch :: XML.Document -> BL.ByteString
 props2patch = XML.renderLBS XML.def . patch . props . fromDocument
    where
-       props cursor = map node (cursor $/ element "{DAV:}response" &/ element "{DAV:}propstat" &/ element "{DAV:}prop" &/ checkName (not . flip elem ["{DAV:}creationdate", "{DAV:}displayname", "{DAV:}getcontentlength", "{DAV:}getcontenttype", "{DAV:}getetag", "{DAV:}getlastmodified", "{DAV:}lockdiscovery", "{DAV:}resourcetype", "{DAV:}supportedlock"]))
+       props cursor = map node (cursor $/ element "{DAV:}response" &/ element "{DAV:}propstat" &/ element "{DAV:}prop" &/ checkName (not . flip elem blacklist))
        patch prop = XML.Document (XML.Prologue [] Nothing []) (root prop) []
-       root prop = XML.Element "D:propertyupdate" [("xmlns:D", "DAV:")]
-           [ XML.NodeElement $ XML.Element "D:set" []
-	     [ XML.NodeElement $ XML.Element "D:prop" [] prop ]
+       root [] = propertyupdate []
+       root prop = propertyupdate
+           [ XML.NodeElement $ XML.Element "D:set" Map.empty
+	     [ XML.NodeElement $ XML.Element "D:prop" Map.empty prop ]
 	   ]
+       propertyupdate = XML.Element "D:propertyupdate" (Map.fromList [("xmlns:D", "DAV:")])
+       blacklist = [ "{DAV:}creationdate"
+                   , "{DAV:}displayname"
+                   , "{DAV:}getcontentlength"
+                   , "{DAV:}getcontenttype"
+                   , "{DAV:}getetag"
+                   , "{DAV:}getlastmodified"
+                   , "{DAV:}lockdiscovery"
+                   , "{DAV:}resourcetype"
+                   , "{DAV:}supportedlock"
+                   ]
+
+getProps :: String -> B.ByteString -> B.ByteString -> IO XML.Document
+getProps url username password = withDS url username password getPropsM
 
 getPropsAndContent :: String -> B.ByteString -> B.ByteString -> IO (XML.Document, (Maybe B.ByteString, BL.ByteString))
 getPropsAndContent url username password = withDS url username password $ do
     getOptions
     o <- get
     when (supportsLocking o) (lockResource True)
-    (do props <- getAllProps
-        body <- getContent
+    (do props <- getPropsM
+        body <- getContentM
         return (props, body)) `finally` when (supportsLocking o) (unlockResource)
 
+putContent :: String -> B.ByteString -> B.ByteString -> (Maybe B.ByteString, BL.ByteString) -> IO ()
+putContent url username password b = withDS url username password $ do
+    getOptions
+    o <- get
+    when (supportsLocking o) (lockResource False)
+    putContentM b `finally` when (supportsLocking o) (unlockResource)
+
 putContentAndProps :: String -> B.ByteString -> B.ByteString -> (XML.Document, (Maybe B.ByteString, BL.ByteString)) -> IO ()
 putContentAndProps url username password (p, b) = withDS url username password $ do
     getOptions
     o <- get
     when (supportsLocking o) (lockResource False)
-    (do putContent b
-        putProps p) `finally` when (supportsLocking o) (unlockResource)
+    (do putContentM b
+        putPropsM p) `finally` when (supportsLocking o) (unlockResource)
+
+deleteContent :: String -> B.ByteString -> B.ByteString -> IO ()
+deleteContent url username password = withDS url username password $ do
+    getOptions
+    o <- get
+    let lock = when (supportsLocking o) (lockResource False)
+    -- a successful delete destroys locks, so only unlock on error
+    let unlock = when (supportsLocking o) (unlockResource)
+    bracketOnError lock (\_ -> unlock) (\_ -> delContentM)
+
+moveContent :: String -> B.ByteString -> B.ByteString -> B.ByteString -> IO ()
+moveContent url newurl username password = withDS url username password $
+    moveContentM newurl
+
+-- | Creates a WebDAV collection, which is similar to a directory.
+--
+-- Returns False if the collection could not be made due to an intermediate
+-- collection not existing. (Ie, collection /a/b/c/d cannot be made until
+-- collection /a/b/c exists.)
+makeCollection :: String -> B.ByteString -> B.ByteString -> IO Bool
+makeCollection url username password = withDS url username password $
+    catchJust
+        (matchStatusCodeException conflict409)
+        (mkCol >> return True)
+        (\_ -> return False)
 
 propname :: XML.Document
 propname = XML.Document (XML.Prologue [] Nothing []) root []
     where
-        root = XML.Element "D:propfind" [("xmlns:D", "DAV:")] [xml|
+        root = XML.Element "D:propfind" (Map.fromList [("xmlns:D", "DAV:")]) [xml|
 <D:allprop>
 |]
 
 locky :: XML.Document
 locky = XML.Document (XML.Prologue [] Nothing []) root []
     where
-        root = XML.Element "D:lockinfo" [("xmlns:D", "DAV:")] [xml|
+        root = XML.Element "D:lockinfo" (Map.fromList [("xmlns:D", "DAV:")]) [xml|
 <D:lockscope>
   <D:exclusive>
 <D:locktype>
